# Copyright 2016 The Fabulous Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    fabulous.widget
    ~~~~~~~~~~~~~~~

    Widget library using terminate.

"""

import os
import math
from datetime import datetime
# import textwrap
from term import stdout, display

class ProgressBar(object):
    """A 3-line progress bar, which looks like::
                                title
        39% [================>----------------------------]
                               message
    
        p = ProgressBar('spam') # create bar
        p.update(0, 'starting spam') # start printing it out
        p.update(50, 'spam almost ready') # progress
        p.update(100, 'spam complete')
    """
    # content, length
    TITLE_FORMAT = {'text':display('bright','cyan') + '%s' + display('default'),
           'length':0,
           'padding':0 }
    BAR_FORMAT = {'text':' %3d%% ' + '[%s'+display('dim')+'%s'+display('default')+']',
           'length':8,
           'padding':2 }
    MESSAGE_FORMAT = {'text': '%s',
           'length': 0,
           'padding': 0 }
        
    def __init__(self, title = None):
        """
        """
        self.drawn = False
        cols = stdout.get_size()[0]
        self.width = cols -1 # TODO: make a better fix for systems that put \n on new line
        self.title = []
        self.barlines = 0
        self.message = []
        self.messageline = None
        self.refresh = False
        self.set_title(title)
    
    def set_title(self, title = None):
        """
        """
        if title == None:
            self.title = []
        else:
            length = self.width - self.TITLE_FORMAT['padding']*2 - self.TITLE_FORMAT['length']
            text = title[:length].center(length) # we need to keep it on one line for now
            padding = ' ' * self.TITLE_FORMAT['padding']
            self.title = [padding + (self.TITLE_FORMAT['text'] % text) + padding]
        self.refresh = self.drawn
            #lines = [(padding + line + padding) for line in textwrap.wrap(
            #          text, self.width - (self.TITLE_FORMAT['padding']*2),
            #          replace_whitespace=False)]
            #self.title = os.linesep.split(
            #          self.TITLE_FORMAT['text'] % os.linesep.join(lines))
    
    def get_title(self):
        """
        """
        return self.title
    
    def get_bar(self, percent):
        """
        """
        barlength = self.width - self.BAR_FORMAT['padding']*2 - self.BAR_FORMAT['length']
        full = int( math.ceil(barlength * (percent / 100.0)) )
        empty = int(barlength - full)
        if full == 0 or empty == 0: fullpiece = ('=' * full)
        else: fullpiece = ('=' * (full-1)) + '>'
        emptypiece = ('-' * empty)
        return [(self.BAR_FORMAT['text'] % (percent, fullpiece, emptypiece))]
    
    def set_message(self, message = None):
        """
        """
        """"""
        if message == None:
            self.message = []
        else:
            length = self.width - self.MESSAGE_FORMAT['padding']*2 - self.MESSAGE_FORMAT['length']
            text = message[:length].center(length) # we need to keep it on one line for now
            padding = ' ' * self.MESSAGE_FORMAT['padding']
            self.message = [padding + (self.MESSAGE_FORMAT['text'] % text) + padding]
    
    def get_message(self):
        """returns None or string"""
        if self.message == []: return None
        else: return os.linesep.join(self.message)
    
    def update(self, percent, message = None, test = False):
        """
        """
        if self.refresh:
            self.clear()
        if self.drawn:
            stdout.move('beginning of line')
            stdout.move('up', len(self.message) + self.barlines)
        else:
            title = self.get_title()
            if title != None:
                for line in self.get_title():
                    stdout.write(line + os.linesep)
            self.drawn = True
        bar = self.get_bar(percent)
        refresh =  (len(bar) != self.barlines)
        self.barlines = len(bar)
        for line in bar:
            stdout.clear('line')
            stdout.write(line)
            stdout.move('down')
            stdout.move('beginning of line')
        if (message != self.get_message()) or refresh:
            stdout.clear('end of screen')
            self.set_message(message)
            for line in self.message:
                stdout.write(line)
                stdout.move('down')
        else: stdout.move('down', len(self.message))
    
    def clear(self):
        """
        """
        if self.drawn:
            stdout.move('beginning of line')
            stdout.move('up', len(self.message))
            stdout.move('up', self.barlines)
            stdout.move('up', len(self.get_title()))
            stdout.clear('end of screen')
            self.drawn = False
        self.refresh = False

class TimedProgressBar(ProgressBar):
    """A 3-line progress bar, which looks like::
                                      title
        39% [================>----------------------------] ETA mm:ss
                                     message
    
        p = ProgressBar('spam') # create bar
        p.update(0, 'starting spam') # start printing it out
        p.update(50, 'spam almost ready') # progress
        p.update(100, 'spam complete')
    """
    
    BAR_FORMAT = {'text':' %3d%% ' + '[%s'+display('dim')+'%s'+display('default')+']',
           'length':13,
           'padding':2 }
    ' ETA 12:23'
    
    # what fraction of percent it acurate too
    precision = 100
    
    def __init__(self, title = None):
        ProgressBar.__init__(self, title)
        self.start = datetime.today()
    
    def get_bar(self, percent):
        now = datetime.today()
        timed = now - self.start
        etatext = ''
        etadiv = int(percent*self.precision)
        if timed.seconds >= 1:
            etatext += ' '
            if int(percent * self.precision) !=0:
                eta = (timed * 100 * self.precision)/int(percent * self.precision)
                days = eta.days
                min, sec = divmod(eta.seconds, 60)
                hours, min = divmod(min, 60)
                if days == 1: etatext += '1 day, '
                elif days: etatext += '%d days, ' % days
                if hours: etatext += '%02d:' % hours
                etatext += '%02d:%02d' % (min, sec)
            else:
                etatext += 'Never'
        barlength = (self.width - self.BAR_FORMAT['padding']*2 
                     - self.BAR_FORMAT['length'] - len(etatext))
        full = int( math.ceil(barlength * (percent / 100.0)) )
        empty = int(barlength - full)
        if full == 0 or empty == 0: fullpiece = ('=' * full)
        else: fullpiece = ('=' * (full-1)) + '>'
        emptypiece = ('-' * empty)
        return [(self.BAR_FORMAT['text'] % (percent, fullpiece, emptypiece))+etatext]

class Spinner(object):
    
    spinners=['/','-','\\','|',]
    
    def __init__(self):
        self.drawn = False
        self.state = 0
    
    def spin(self):
        if self.drawn == True: self.clear()
        else: self.drawn = True
        stdout.write(self.spinners[self.state])
        self.state = (self.state + 1) % len(self.spinners)
        
    def clear(self):
        stdout.clear('left')
        stdout.move('left')



# Copyright 2016 The Fabulous Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A series of functions that provide useful casts from strings into other common objects.

Mainly for usage with input_object in the above package.
"""

from os import path

def yes_no(value):
    """For a yes or no question, returns a boolean.
    """
    if value.lower() in ('yes','y'):
        return True
    if value.lower() in ('no','n'):
        return False
    raise ValueError("value should be 'yes' or 'no'")


def file(value, **kwarg):
    """value should be a path to file in the filesystem.
    
    returns a file object
    """
    #a bit weird, but I don't want to hard code default values
    try:
        f = open(value, **kwarg)
    except IOError as e:
        raise ValueError("unable to open %s : %s" % (path.abspath(value), e))
    return f



# -*- coding: utf-8 -*-
#
# Copyright 2016 The Fabulous Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    fabulous.text
    ~~~~~~~~~~~~~

    The text module makes it possible to print TrueType text to the terminal.
    This functionality is available on the command line::

        jart@compy:~$ fabulous-text --help
        jart@compy:~$ fabulous-text --skew=5 --shadow 'Fabulous!'
        jart@compy:~$ python -m fabulous.text --help

    Or as a Python library:

    .. code-block:: python

        from fabulous import text
        print text.Text("Fabulous!", color='#0099ff', shadow=True, skew=5)

    To make things simple, Fabulous bundles the following Google Noto Fonts
    which look good and are guaranteed to work no matter what:

    - NotoSans-Bold
    - NotoEmoji-Regular

    For other fonts, Fabulous will do its best to figure out where they are
    stored. If Fabulous has trouble finding your font, try using an absolute
    path *with* the extension. It's also possible to put the font in the
    ``~/.fonts`` directory and then running ``fc-cache -fv ~/.fonts``.

    You can run ``fabulous-text --list`` to see what fonts are available.

"""

from __future__ import print_function

import os
import sys

from fabulous import utils, image, grapefruit
from fabulous.compatibility import printy

try:
    unicode = unicode
except NameError:
    unicode = str
    basestring = (str, bytes)


class Text(image.Image):
    u"""Renders TrueType Text to Terminal

    I'm a sub-class of :class:`fabulous.image.Image`.  My job is
    limited to simply getting things ready.  I do this by:

    - Turning your text into an RGB-Alpha bitmap image using
      :mod:`PIL`

    - Applying way cool effects (if you choose to enable them)

    For example::

        >>> assert Text("Fabulous", shadow=True, skew=5)

        >>> txt = Text("lorem ipsum", font="NotoSans-Bold")
        >>> len(str(txt)) > 0
        True
        >>> txt = Text(u"😃", font="NotoSans-Bold")
        >>> len(str(txt)) > 0
        True

    :param text:   The text you want to display as a string.

    :param fsize:  The font size in points.  This obviously end up
                   looking much larger because in fabulous a single
                   character is treated as one horizontal pixel and two
                   vertical pixels.

    :param color:  The color (specified as you would in HTML/CSS) of
                   your text.  For example Red could be specified as
                   follows: ``red``, ``#00F`` or ``#0000FF``.

    :param shadow: If true, render a simple drop-shadow beneath text.
                   The Fabulous logo uses this feature.

    :param skew:   Skew size in pixels.  This applies an affine
                   transform to shift the top-most pixels to the right.
                   The Fabulous logo uses a five pixel skew.

    :param font:   The TrueType font you want.  If this is not an
                   absolute path, Fabulous will search for your font by
                   globbing the specified name in various directories.
    """

    def __init__(self, text, fsize=23, color="#0099ff", shadow=False,
                 skew=None, font='NotoSans-Bold'):
        utils.pil_check()
        from PIL import Image, ImageFont, ImageDraw
        self.text = text
        self.color = grapefruit.Color.NewFromHtml(color)
        self.font = ImageFont.truetype(resolve_font(font), fsize)
        skew = skew or 0
        size = tuple([n + 3 + skew for n in self.font.getsize(self.text)])
        self.img = Image.new("RGBA", size, (0, 0, 0, 0))
        cvs = ImageDraw.Draw(self.img)
        if shadow:
            cvs.text((2 + skew, 2), self.text,
                     font=self.font,
                     fill=(150, 150, 150, 150))
        cvs.text((1 + skew, 1), self.text,
                 font=self.font,
                 fill=self.color.html)
        if skew:
            self.img = self.img.transform(
                size, Image.AFFINE, (1.0, 0.1 * skew, -1.0 * skew,
                                     0.0, 1.0, 0.0))
        self.resize(None)


class FontNotFound(ValueError):
    """I get raised when the font-searching hueristics fail

    This class extends the standard :exc:`ValueError` exception so you
    don't have to import me if you don't want to.
    """


def resolve_font(name):
    """Turns font names into absolute filenames

    This is case sensitive. The extension should be omitted.

    For example::

        >>> path = resolve_font('NotoSans-Bold')

        >>> fontdir = os.path.join(os.path.dirname(__file__), 'fonts')
        >>> noto_path = os.path.join(fontdir, 'NotoSans-Bold.ttf')
        >>> noto_path = os.path.abspath(noto_path)
        >>> assert path == noto_path

    Absolute paths are allowed::

        >>> resolve_font(noto_path) == noto_path
        True

    Raises :exc:`FontNotFound` on failure::

        >>> try:
        ...     resolve_font('blahahaha')
        ...     assert False
        ... except FontNotFound:
        ...     pass

    """
    if os.path.exists(name):
        return os.path.abspath(name)
    fonts = get_font_files()
    if name in fonts:
        return fonts[name]
    raise FontNotFound("Can't find %r :'(  Try adding it to ~/.fonts" % name)

font_roots = [
    '/usr/share/fonts/truetype',                # where ubuntu puts fonts
    '/usr/share/fonts',                         # where fedora puts fonts
    os.path.expanduser('~/.local/share/fonts'), # custom user fonts
    os.path.expanduser('~/.fonts'),             # custom user fonts
    os.path.abspath(os.path.join(os.path.dirname(__file__), 'fonts')),
]

@utils.memoize
def get_font_files():
    """Returns a list of all font files we could find

    Returned as a list of dir/files tuples::

        get_font_files() -> {'FontName': '/abs/FontName.ttf', ...]

    For example::

        >>> fonts = get_font_files()
        >>> 'NotoSans-Bold' in fonts
        True
        >>> fonts['NotoSans-Bold'].endswith('/NotoSans-Bold.ttf')
        True

    """
    result = {}
    for root in font_roots:
        for path, dirs, names in os.walk(root):
            for name in names:
                if name.endswith(('.ttf', '.otf')):
                    result[name[:-4]] = os.path.join(path, name)
    return result


def main():
    """Main function for :command:`fabulous-text`."""
    import optparse
    parser = optparse.OptionParser()
    parser.add_option(
        "-l", "--list", dest="list", action="store_true", default=False,
        help=("List available fonts"))
    parser.add_option(
        "-S", "--skew", dest="skew", type="int", default=None,
        help=("Apply skew effect (measured in pixels) to make it look "
              "extra cool.  For example, Fabulous' logo logo is skewed "
              "by 5 pixels.  Default: %default"))
    parser.add_option(
        "-C", "--color", dest="color", default="#0099ff",
        help=("Color of your text.  This can be specified as you would "
              "using HTML/CSS.  Default: %default"))
    parser.add_option(
        "-B", "--term-color", dest="term_color", default=None,
        help=("If you terminal background isn't black, please change "
              "this value to the proper background so semi-transparent "
              "pixels will blend properly."))
    parser.add_option(
        "-F", "--font", dest="font", default='NotoSans-Bold',
        help=("Name of font file, or absolute path to one. Use the --list "
              "flag to see what fonts are available. Fabulous bundles the "
              "NotoSans-Bold and NotoEmoji-Regular fonts, which are guaranteed "
              "to work. Default: %default"))
    parser.add_option(
        "-Z", "--size", dest="fsize", type="int", default=23,
        help=("Size of font in points.  Default: %default"))
    parser.add_option(
        "-s", "--shadow", dest="shadow", action="store_true", default=False,
        help=("Size of font in points.  Default: %default"))
    (options, args) = parser.parse_args(args=sys.argv[1:])
    if options.list:
        print("\n".join(sorted(get_font_files())))
        return
    if options.term_color:
        utils.term.bgcolor = options.term_color
    text = " ".join(args)
    if not isinstance(text, unicode):
        text = text.decode('utf-8')
    for line in text.split("\n"):
        fab_text = Text(line, skew=options.skew, color=options.color,
                        font=options.font, fsize=options.fsize,
                        shadow=options.shadow)
        for chunk in fab_text:
            printy(chunk)


if __name__ == '__main__':
    main()



# Copyright 2016 The Fabulous Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import with_statement

import time
import curses


class Canvas(object):
    def __init__(self, encoding='UTF-8'):
        self.encoding = encoding

    def __enter__(self):
        self.win = curses.initscr()
        curses.start_color()
        curses.init_color(200, 1000, 300, 0)
        curses.init_pair(1, 200, curses.COLOR_WHITE)
        return self

    def __exit__(self, type_, value, traceback):
        curses.endwin()

    def __setitem__(self, xy, val):
        self.win.attron(curses.color_pair(1))
        (x, y) = xy
        self.win.addch(x, y, val)


if __name__ == '__main__':
    import locale
    locale.setlocale(locale.LC_ALL, '')
    encoding = locale.getpreferredencoding()
    with Canvas(encoding=encoding) as canvas:
        canvas[5, 5] = 'Y'
        canvas.win.refresh()
        time.sleep(5.0)
